#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Convert LEMI data and metadata into day-long mseed and StationXML files.

Maeva Pourpoint - IRIS/PASSCAL
"""

import argparse
import re
import sys

from pathlib import Path
from PySide2 import QtWidgets
from typing import Any, Dict, Optional, Tuple

from lemi2seed.data_conversion import write_daylong_mseed, write_log_file
from lemi2seed.lemi_data import LemiData
from lemi2seed.gui_view import LemiWindow, VALID_NUM_E_PAIRS
from lemi2seed.lemi_metadata import LemiMetadata
from lemi2seed.logging import initiate_logger, parse_config_ini, setup_logger
from lemi2seed.utils import is_empty

# Read config.ini file and set up logging
config = parse_config_ini()
logger = setup_logger(__name__)
initiate_logger(logger, config['software']['version'])

VALID_CHA_PORTS = ['E1', 'E2', 'E3', 'E4']
VALID_DIRS = {1: ['North-South', 'Ex'], 2: ['East-West', 'Ey']}


def setup_args() -> argparse.ArgumentParser:
    """Set up command line interface."""
    parser = argparse.ArgumentParser(prog='lemi2seed',
                                     formatter_class=argparse.RawTextHelpFormatter,
                                     description='''
Convert LEMI data and metadata to day-long mseed and StationXML files.
lemi2seed can run under three distinct modes:
- quality control mode ('lemi2seed -d path2data -qc')
- gui only mode ('lemi2seed -d path2data -g)
- field sheet and gui mode (lemi2seed -d path2data -m path2metadata)
The 'quality control' mode allows to quickly convert LEMI data to mseed. The GUI and handling of metadata are bypassed.
In the 'gui only' mode, metadata are entered/modified using only the GUI.
In the 'field sheet and gui' mode, metadata are prepopulated from parsing the user-provided field sheets and modified using the GUI.''')   # noqa: E501
    parser.add_argument('-d',
                        dest='path2data',
                        metavar='path2data',
                        help='''Path to the station directory where all folders/files generated by the LEMI-424 are stored''')   # noqa: E501
    parser.add_argument('-m',
                        dest='path2md',
                        metavar='path2md',
                        help='''Path to the directory where all LEMI field spreadsheets are stored''')
    parser.add_argument('-s',
                        dest='path2savedmd',
                        metavar='path2savedmd',
                        help='''Path to the .pkl file where metadata were previously saved''')
    parser.add_argument('-od',
                        dest='output_mseed',
                        metavar='output_mseed',
                        default='MSEED',
                        help='''Output directory name for day-long mseed files''')
    parser.add_argument('-om',
                        dest='output_xml',
                        metavar='output_xml',
                        default='STATIONXML',
                        help='''Output directory name for StationXML files''')
    parser.add_argument('-ol',
                        dest='output_log',
                        metavar='output_log',
                        default='LOGS',
                        help='''Output directory name for calibration coefficient information''')
    parser.add_argument('-qc',
                        dest='qc_mode',
                        action='store_true',
                        default=False,
                        help='''Runs lemi2seed in quality control mode. Handling of metadata is bypassed.''')
    parser.add_argument('-g',
                        dest='gui_only_mode',
                        action='store_true',
                        default=False,
                        help='''Runs lemi2seed in gui only mode. All metadata are entered using the GUI
(bypassing the use of field sheets).''')
    return parser


def get_args() -> Tuple[Path, Optional[Path], Optional[Path], bool, bool, Path, Path, Optional[Path]]:
    """
    Parse command line arguments and perform a serie of checks on command line
    arguments.
    """
    # Set up command line interface and parse command line arguments
    parser = setup_args()
    args = parser.parse_args()

    # Path to LEMI ASCII files
    if args.path2data and Path(args.path2data).is_dir():
        path2data = Path(args.path2data).resolve()
    else:
        logger.error("Path to LEMI data not provided! Required input! "
                     "Try --help option.")
        sys.exit(1)

    # Running modes, path to LEMI field sheets and path to .pkl file
    qc_only = args.qc_mode
    gui_only = args.gui_only_mode
    if all([qc_only, gui_only]):
        logger.error("lemi2seed cannot run under both the 'quality control' "
                     "and 'gui only' mode at the same time! Try --help option "
                     "or choose one or the other mode.")
        sys.exit(1)

    if args.path2md is None:
        if not any([qc_only, gui_only]):
            logger.error("No metadata path provided. Required input when "
                         "running in 'field sheet and gui' mode! Try --help "
                         "option.")
            sys.exit(1)
        path2md = args.path2md
    else:
        if Path(args.path2md).is_dir():
            if qc_only:
                logger.info("Handling of the field sheets is bypassed when "
                            "running in 'quality control' mode.")
                path2md = None
            elif gui_only:
                logger.info("All metadata should be entered using the GUI in "
                            "'gui only' mode. Bypassing the use of field sheets.")
                path2md = None
            else:
                path2md = Path(args.path2md).resolve()
        else:
            logger.error("The path to the LEMI field spreadsheets does not "
                         "exist!")
            if any([qc_only, gui_only]):
                logger.info("In 'quality control' or 'gui only' mode, field "
                            "sheets are not parsed.")
                path2md = None
            else:
                logger.error("A correct metadata path is a required input "
                             "when running in 'field sheet and gui' mode!")
                sys.exit(1)

    if args.path2savedmd is not None:
        if qc_only:
            logger.info("Handling of previously saved metadata is bypassed when "
                        "running in 'quality control' mode.")
            path2savedmd = None
        elif gui_only:
            if not Path(args.path2savedmd).is_file():
                logger.error("The provided path to a previously saved metadata "
                             "file (*.pkl) does not exist! Bypassing loading "
                             "of metadata fields before launching the GUI.")
                path2savedmd = None
            else:
                path2savedmd = Path(args.path2savedmd).resolve()
        else:
            logger.info("The loading of a previously saved metadata file (*.pkl) "
                        "is bypassed when running in 'field sheet and gui' "
                        "mode. Try --help option for more information.")
            path2savedmd = None
    else:
        path2savedmd = args.path2savedmd

    # Output directories
    output_mseed = Path(args.output_mseed).resolve()
    check_output_dir(output_mseed, 'day-long mseed')
    output_log = Path(args.output_log).resolve()
    check_output_dir(output_log, 'log')
    if qc_only:
        output_xml = None
    else:
        output_xml = Path(args.output_xml).resolve()
        check_output_dir(output_xml, 'StationXML')

    return path2data, path2md, path2savedmd, qc_only, gui_only, output_mseed, output_log, output_xml


def check_output_dir(output_dir: Path, output_type: str) -> None:
    """
    Check whether the user-defined output directory exists.
    If not, create directory.
    """
    if not output_dir.exists():
        try:
            output_dir.mkdir(parents=True)
        except OSError:
            logger.exception("Failed to create output directory name for "
                             "{} files - {}".format(output_type, output_dir))
            sys.exit(1)


def get_net_inputs() -> str:
    """Request user to enter 'network code'."""
    while True:
        net = input("Enter network code: ")
        try:
            valid = re.match(r"^\w{2}$", net)
        except TypeError:
            logger.error("The network code should be a string.")
            continue
        if valid is None:
            logger.error("The network code should be two alphanumeric "
                         "character long.")
            continue
        else:
            break
    return net


def get_sta_inputs() -> str:
    """Request user to enter 'station name'."""
    while True:
        sta = input("Enter station name: ")
        try:
            valid = re.match(r"^\w{3,5}$", sta)
        except TypeError:
            logger.error("The station name should be a string.")
            continue
        if valid is None:
            logger.error("The station name should be between three and five "
                         "alphanumeric character long.")
            continue
        else:
            break
    return sta


def get_num_inputs() -> int:
    """Request user to provide number of electrode pairs installed."""
    options = ', '.join([str(x) for x in VALID_NUM_E_PAIRS])
    while True:
        try:
            nums = int(input("Enter number of electrode pairs installed (valid "
                             "options: {}):".format(options)))
        except ValueError:
            logger.error("Please enter appropriate number of electrode pairs "
                         "(integer expected)!")
            continue
        if nums not in VALID_NUM_E_PAIRS:
            logger.error("Please enter appropriate number of electrode pairs "
                         "(valid options: {})!".format(options))
            continue
        else:
            break
    return nums


def get_cha_port_inputs(num: int) -> str:
    """
    Request user to provide channel port to which each electrode pair was
    connected.
    """
    options = ', '.join(VALID_CHA_PORTS)
    while True:
        cha_port = input("Enter channel port for electrode pair {} (valid "
                         "options: {}): ".format(num, options))
        if cha_port not in VALID_CHA_PORTS:
            logger.error("Please enter appropriate channel port number "
                         "(valid options: {})!".format(options))
            continue
        else:
            break
    return cha_port


def get_dir_inputs(num: int) -> int:
    """Request user to provide direction of each electrode pair."""
    options = ' & '.join(["{} for {}".format(k, v[0]) for k, v in VALID_DIRS.items()])
    while True:
        try:
            cha_dir = int(input("Enter direction of electrode pair {} ({}): "
                                .format(num, options)))
        except ValueError:
            logger.error("Please enter appropriate channel direction "
                         "(integer expected)!")
            continue
        if cha_dir not in VALID_DIRS.keys():
            logger.error("Please enter appropriate channel direction "
                         "(valid options: {})!".format(options))
            continue
        else:
            break
    return cha_dir


def get_elec_inputs(user_inputs: Dict) -> Dict:
    """
    Request user to provide following iputs:
    - number of electrode pairs installed
    - channel port to which each electrode pair was connected (if any)
    - direction of each electrode pair (if any)
    """
    nums = get_num_inputs()
    for num in range(1, nums + 1):
        cha_port = get_cha_port_inputs(num)
        user_inputs['e_chas'].append(cha_port)
        cha_dir = get_dir_inputs(num)
        user_inputs['e_info'][cha_port] = VALID_DIRS[cha_dir][1]
        VALID_CHA_PORTS.remove(cha_port)
    return user_inputs


def get_qc_inputs() -> Dict:
    """
    Request user inputs via the command line. Only use in QC mode.
    User inputs:
    - network code (key: net)
    - station name (key: sta)
    - number of electrode pairs installed
    - channel port used for each electrode pair (key: e_chas)
    - direction of each electrode pair (key: e_info)
    are stored in a dictionary.
    """
    logger.info("Please provide the following metadata before proceeding with "
                "the data conversion to day-long mseed files.")
    user_inputs: Dict[str, Any] = {}
    user_inputs['net'] = get_net_inputs()
    user_inputs['sta'] = get_sta_inputs()
    user_inputs['e_chas'] = []
    user_inputs['e_info'] = {}
    user_inputs = get_elec_inputs(user_inputs)
    return user_inputs


def reformat_user_inputs(inputs: Dict) -> str:
    """
    Reformat qc input parameters in a user-friendly format. Reformatted inputs
    will be saved in the msg file and displayed in the command line.
    """
    num_e_pairs = len(inputs['e_chas'])
    msg1 = (f"  Network code: {inputs['net']}\n"
            f"  Station name: {inputs['sta']}\n"
            f"  Number of electrode pairs installed: {num_e_pairs}\n")
    msg2 = ""
    for i in range(num_e_pairs):
        cha_port = inputs['e_chas'][i]
        comp = inputs['e_info'][cha_port]
        dir = 'North-South' if comp == 'Ex' else 'East-West'
        msg2 += (f"  Channel port number for electrode pair {i+1}: {cha_port}\n"
                 f"  Direction of electrode pair {i+1}: {dir}\n")
    msg = msg1 + msg2
    return msg.strip("\n")


def launch_gui(lemi_data: LemiData, lemi_md: LemiMetadata, path2savedmd: Path) -> None:
    """
    Open GUI window, set the model data structure and handle metadata
    processing and archival steps.
    Only use in non-QC mode.
    """
    app = QtWidgets.QApplication(sys.argv)
    lemi2seed_gui = LemiWindow(lemi_data, lemi_md)
    if path2savedmd:
        lemi2seed_gui.load_md(path2savedmd)
    lemi2seed_gui.show()
    sys.exit(app.exec_())


def main():
    # Read command line arguments
    path2data, path2md, path2savedmd, qc_only, gui_only, output_mseed, output_log, output_xml = get_args()

    # Parse data files
    lemi_data = LemiData(path2data, output_mseed, output_log)
    lemi_data.prep_data()

    # Convert recorded data to day-long mseed
    if qc_only:
        logger.info("Running lemi2seed in 'quality control' mode.")
        qc_inputs = get_qc_inputs()
        logger.info("Summary of entered input parameters:\n{}"
                    .format(reformat_user_inputs(qc_inputs)))
        lemi_data.update_data(qc_inputs=qc_inputs)
        logger.info("Writing day-long mseed files.")
        write_daylong_mseed(lemi_data)
        if lemi_data.def_files:
            write_log_file(lemi_data)
    # (Optional) Parse field sheets and prepopulate metadata in GUI
    # Populate metadata in GUI framework
    # Write day-long mseed and StationXML files after:
    # - all required metadata have been populated
    # - validation of populated metadata
    else:
        lemi_md = LemiMetadata(path2md, output_xml, lemi_data)
        if not gui_only and not is_empty(lemi_md.filenames):
            logger.info("Running lemi2seed in 'field sheet and gui' mode.")
            md_fields = lemi_md.from_field_sheets
            if md_fields is not None:
                logger.info("Pre-populating the GUI metadata fields using the "
                            "information provided in the field sheet(s).")
                lemi_md.populate_props(md_fields)
        else:
            logger.info("Running lemi2seed in 'gui only' mode.")
        logger.info("Launching the GUI.")
        launch_gui(lemi_data, lemi_md, path2savedmd)


if __name__ == '__main__':
    main()
